from django.db import transaction
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import build_tree


class TestElementNeighbors(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug="volume")
        cls.page_type = cls.corpus.types.get(slug="page")

    def test_element_neighbors_acl(self):
        """
        A Guest access is required to list neighbors of an element
        """
        private_corpus = Corpus.objects.create(name="Private", public=False)
        private_type = private_corpus.types.create(slug="folder")
        elements = build_tree(
            {
                "B": "A",
                "C": "A",
                "D": "A",
            },
            corpus=private_corpus,
            type=private_type,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:elements-neighbors", kwargs={"pk": str(elements["A"].id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have a read access to this element."})

    def test_element_neighbors(self):
        r"""
        A non authenticated user is able to list neighbors of a public element
            Z
            |
          X Y
         /|X \
        B A C D
        """
        elements = build_tree(
            {
                "Y": "Z",
                "B": "X",
                "A": ["X", "Y"],
                "C": "X",
                "D": "Y",
            },
            corpus=self.corpus,
            type=self.volume_type,
        )

        response = self.client.get(reverse("api:elements-neighbors", kwargs={"pk": str(elements["A"].id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json(),
            sorted([{
                # 2 neighbors via X (B, A, C) A has position 1
                "ordering": 1,
                "previous": {
                    "id": str(elements["B"].id),
                    "type": "volume",
                    "name": elements["B"].name
                },
                "next": {
                    "id": str(elements["C"].id),
                    "type": "volume",
                    "name": elements["C"].name
                },
                "path": [{
                    "id": str(elements["X"].id),
                    "type": "volume",
                    "name": elements["X"].name
                }],
            }, {
                # 1 neighbors via Y (A, D) A has position 0
                "ordering": 0,
                "previous": None,
                "next": {
                    "id": str(elements["D"].id),
                    "type": "volume",
                    "name": elements["D"].name
                },
                "path": [{
                    "id": str(elements["Z"].id),
                    "type": "volume",
                    "name": elements["Z"].name
                }, {
                    "id": str(elements["Y"].id),
                    "type": "volume",
                    "name": elements["Y"].name
                }]
            }],
                # Order result by paths, as we can't predict the test generated uuids
                key=lambda path: "".join([p["id"] for p in path["path"]]))
        )

    def test_element_neighbors_missing_ordering(self):
        """
        Check ElementNeighbors handles non-sequential orderings
        """
        elements = build_tree(
            {
                "A": "W",
                "B": "W",
                "C": "W",
            },
            corpus=self.corpus,
            type=self.volume_type,
        )
        c_path = elements["C"].paths.get()
        c_path.ordering = 6584

        # The unique_element_orderings trigger will fail because an element is not supposed to have
        # multiple orderings within the same parent. This trigger is deferred, meaning it will only run
        # after the test ends, during teardown, so we cannot ignore it with pgtrigger.ignore().
        # Instead, we use a savepoint that we rollback immediately after the request, so that the trigger never runs.
        sid = transaction.savepoint()
        try:
            c_path.save()
            response = self.client.get(reverse("api:elements-neighbors", kwargs={"pk": str(elements["B"].id)}))
        finally:
            transaction.savepoint_rollback(sid)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json(),
            [{
                "ordering": 1,
                "previous": {
                    "id": str(elements["A"].id),
                    "type": "volume",
                    "name": elements["A"].name
                },
                "next": {
                    "id": str(elements["C"].id),
                    "type": "volume",
                    "name": elements["C"].name
                },
                "path": [{
                    "id": str(elements["W"].id),
                    "type": "volume",
                    "name": elements["W"].name
                }]
            }]
        )

    def test_sub_neighbors(self):
        r"""
        Ensure the children of some elements of the same type do not interfere with the neighboring elements
           V
          /|\
         A B C
          / \
         B1 B2
        """
        elements = build_tree(
            {
                "A": "V",
                "B": "V",
                "C": "V",
                "B1": "B",
                "B2": "B",
            },
            corpus=self.corpus,
            type=self.volume_type,
        )
        response = self.client.get(reverse("api:elements-neighbors", kwargs={"pk": str(elements["B"].id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            response.json(),
            [{
                "ordering": 1,
                "previous": {
                    "id": str(elements["A"].id),
                    "type": "volume",
                    "name": elements["A"].name
                },
                "next": {
                    "id": str(elements["C"].id),
                    "type": "volume",
                    "name": elements["C"].name
                },
                "path": [{
                    "id": str(elements["V"].id),
                    "type": "volume",
                    "name": elements["V"].name
                }]
            }]
        )

    def test_not_found(self):
        self.assertFalse(Element.objects.filter(id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa").exists())
        response = self.client.get(reverse(
            "api:elements-neighbors",
            kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"},
        ))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
