import base64
import hashlib
import random
import uuid
from io import BytesIO
from pathlib import Path
from textwrap import dedent
from unittest.mock import PropertyMock, call, patch, seal

from botocore.exceptions import ClientError
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat, load_pem_private_key
from django.conf import settings
from django.core import mail
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.api import timezone as api_tz
from arkindex.ponos.authentication import AgentUser
from arkindex.ponos.models import ACTIVE_STATES, FINAL_STATES, GPU, Agent, Farm, Secret, State, Task, encrypt
from arkindex.ponos.tasks import notify_process_completion
from arkindex.process.models import Process, ProcessMode, Revision, WorkerActivityState, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import build_public_key
from arkindex.users.models import Right, Role, User
from rest_framework_simplejwt.tokens import AccessToken, RefreshToken

PONOS_PRIVATE_KEY = Path(__file__).absolute().parent / "fixtures" / "ponos.key"


# Helper to format a datetime as output by DRF
def str_date(d):
    return d.isoformat().replace("+00:00", "Z")


@override_settings(PONOS_LOG_TAIL=42)
class TestAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page_1 = cls.corpus.elements.get(name="Volume 1, page 1r")
        cls.page_2 = cls.corpus.elements.get(name="Volume 1, page 2r")
        cls.default_farm = Farm.objects.create(name="Default farm")
        cls.wheat_farm = Farm.objects.get(name="Wheat farm")
        pubkey = build_public_key()
        cls.agent = AgentUser.objects.create(
            id=uuid.UUID(hashlib.md5(pubkey.encode("utf-8")).hexdigest()),
            farm=cls.wheat_farm,
            hostname="ghostname",
            cpu_cores=2,
            cpu_frequency=1e9,
            public_key=pubkey,
            ram_total=2e9,
            last_ping=timezone.now(),
            cpu_load=.1,
            ram_load=.1e9,
        )
        cls.rev = Revision.objects.first()
        cls.process = Process.objects.get(mode=ProcessMode.Workers)
        cls.process.run()
        cls.task1, cls.task2, cls.task3 = cls.process.tasks.all()

        # Brand new user and corpus with no preexisting rights
        new_user = User.objects.create(email="another@user.com")
        new_corpus = Corpus.objects.create(name="corpus the second")
        Right.objects.create(user=new_user, content_object=new_corpus, level=50)
        cls.process2 = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=new_user,
            corpus=new_corpus,
            farm=cls.default_farm,
        )
        cls.process2.run()
        cls.task4 = cls.process2.tasks.first()

        cls.dla = WorkerVersion.objects.get(worker__slug="dla")
        cls.recognizer = WorkerVersion.objects.get(worker__slug="reco")
        cls.gpu1 = cls.agent.gpus.create(
            id="108c6524-c63a-4811-bbed-9723d32a0688",
            name="GPU1",
            index=0,
            ram_total=2 * 1024 * 1024 * 1024,
        )
        cls.gpu2 = cls.agent.gpus.create(
            id="f30d1407-92bb-484b-84b0-0b8bae41ca91",
            name="GPU2",
            index=1,
            ram_total=8 * 1024 * 1024 * 1024,
        )

    def test_task_details_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    @patch("arkindex.project.aws.s3")
    def test_task_details_own_task(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        with self.assertNumQueries(3):
            resp = self.client.get(
                reverse("api:task-details", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            resp.json(),
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "slug": "initialisation",
                "state": "unscheduled",
                "parents": [],
                "logs": "Failed successfully",
                "full_log": "http://somewhere",
                "extra_files": {},
                "agent": None,
                "gpu": None,
                "shm_size": None,
            },
        )

        self.assertEqual(s3_mock.Object.call_count, 2)
        self.assertEqual(s3_mock.Object().get.call_count, 1)
        self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
        self.assertEqual(s3_mock.meta.client.generate_presigned_url.call_count, 1)
        self.assertEqual(
            s3_mock.meta.client.generate_presigned_url.call_args,
            call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
        )

    def test_task_details_other_task_no_access_forbidden(self):
        # Make corpus private just for this test
        self.corpus.public = False
        self.corpus.save()

        with self.assertNumQueries(2):
            resp = self.client.get(
                reverse("api:task-details", args=[self.task2.id]),
                HTTP_AUTHORIZATION=f"Ponos {self.task4.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_task_details_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_task_details_requires_process_guest(self):
        self.process.creator = self.superuser
        self.process.save()
        self.corpus.memberships.filter(user=self.user).delete()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.aws.s3")
    def test_task_details_process_level_corpus(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.process.creator = self.superuser
        self.process.save()
        self.client.force_login(self.user)

        for role in Role:
            s3_mock.reset_mock()
            s3_mock.reset_mock()
            # Recreate the BytesIO each time, because its contents get consumed each time the API is called
            s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}

            with self.subTest(role=role):
                self.corpus.memberships.filter(user=self.user).update(level=role.value)

                with self.assertNumQueries(6):
                    resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
                    self.assertEqual(resp.status_code, status.HTTP_200_OK)

                self.assertDictEqual(
                    resp.json(),
                    {
                        "id": str(self.task1.id),
                        "run": 0,
                        "depth": 0,
                        "slug": "initialisation",
                        "state": "unscheduled",
                        "parents": [],
                        "logs": "Failed successfully",
                        "full_log": "http://somewhere",
                        "extra_files": {},
                        "agent": None,
                        "gpu": None,
                        "shm_size": None,
                    },
                )

                self.assertEqual(s3_mock.Object.call_count, 2)
                self.assertEqual(s3_mock.Object().get.call_count, 1)
                self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
                self.assertEqual(s3_mock.meta.client.generate_presigned_url.call_count, 1)
                self.assertEqual(
                    s3_mock.meta.client.generate_presigned_url.call_args,
                    call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
                )

    @patch("arkindex.project.aws.s3")
    def test_task_details_process_level_repo(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.client.force_login(self.user)
        self.process.mode = ProcessMode.Repository
        self.process.corpus = None
        self.process.revision = self.rev
        self.process.save()
        membership = self.rev.repo.memberships.create(user=self.user, level=Role.Guest.value)

        for role in Role:
            s3_mock.reset_mock()
            s3_mock.reset_mock()
            # Recreate the BytesIO each time, because its contents get consumed each time the API is called
            s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}

            with self.subTest(role=role):
                membership.level = role.value
                membership.save()

                with self.assertNumQueries(7):
                    resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
                    self.assertEqual(resp.status_code, status.HTTP_200_OK)

                self.assertDictEqual(
                    resp.json(),
                    {
                        "id": str(self.task1.id),
                        "run": 0,
                        "depth": 0,
                        "slug": "initialisation",
                        "state": "unscheduled",
                        "parents": [],
                        "logs": "Failed successfully",
                        "full_log": "http://somewhere",
                        "extra_files": {},
                        "agent": None,
                        "gpu": None,
                        "shm_size": None,
                    },
                )

                self.assertEqual(s3_mock.Object.call_count, 2)
                self.assertEqual(s3_mock.Object().get.call_count, 1)
                self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
                self.assertEqual(s3_mock.meta.client.generate_presigned_url.call_count, 1)
                self.assertEqual(
                    s3_mock.meta.client.generate_presigned_url.call_args,
                    call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
                )

    @patch("arkindex.project.aws.s3")
    def test_task_details(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            resp.json(),
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "slug": "initialisation",
                "state": "unscheduled",
                "parents": [],
                "logs": "Failed successfully",
                "full_log": "http://somewhere",
                "extra_files": {},
                "agent": None,
                "gpu": None,
                "shm_size": None,
            },
        )

        self.assertEqual(s3_mock.Object.call_count, 2)
        self.assertEqual(s3_mock.Object().get.call_count, 1)
        self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
        self.assertEqual(s3_mock.meta.client.generate_presigned_url.call_count, 1)
        self.assertEqual(
            s3_mock.meta.client.generate_presigned_url.call_args,
            call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
        )

    def test_task_definition(self):
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:task-definition", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.task1.id),
            "slug": "initialisation",
            "parents": [],
            "image": settings.ARKINDEX_TASKS_IMAGE,
            "command": f"python -m arkindex_tasks.init_elements {self.process.id} --chunks-number 1",
            "shm_size": None,
            "env": {
                "ARKINDEX_API_CSRF_COOKIE": "arkindex.csrf",
                "ARKINDEX_API_TOKEN": "deadbeefTestToken",
                "ARKINDEX_API_URL": "http://localhost:8000/api/v1/",
                "ARKINDEX_CORPUS_ID": str(self.corpus.id),
                "ARKINDEX_PROCESS_ID": str(self.process.id),
                "ARKINDEX_TASK_TOKEN": self.task1.token,
            },
            "has_docker_socket": False,
            "image_artifact_url": None,
            "agent_id": None,
            "gpu_id": None,
            "process_id": str(self.process.id),
            "extra_files": {},
            "s3_logs_put_url": None,
            "state": State.Unscheduled.value,
        })

    def test_task_definition_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:task-definition", args=[self.task1.id]))
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_task_definition_requires_agent(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:task-definition", args=[self.task1.id]))
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_task_definition_task_forbidden(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:task-definition", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_task_definition_image_artifact(self):
        self.task1.image_artifact = self.task1.artifacts.create(
            path="path/to/image.tar",
            size=1234567,
            content_type="application/x-tar",
        )
        self.task1.save()

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:task-definition", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.task1.id),
            "slug": "initialisation",
            "parents": [],
            "image": settings.ARKINDEX_TASKS_IMAGE,
            "command": f"python -m arkindex_tasks.init_elements {self.process.id} --chunks-number 1",
            "shm_size": None,
            "env": {
                "ARKINDEX_API_CSRF_COOKIE": "arkindex.csrf",
                "ARKINDEX_API_TOKEN": "deadbeefTestToken",
                "ARKINDEX_API_URL": "http://localhost:8000/api/v1/",
                "ARKINDEX_CORPUS_ID": str(self.corpus.id),
                "ARKINDEX_PROCESS_ID": str(self.process.id),
                "ARKINDEX_TASK_TOKEN": self.task1.token,
            },
            "has_docker_socket": False,
            "image_artifact_url": f"http://testserver/api/v1/task/{str(self.task1.id)}/artifact/path/to/image.tar",
            "agent_id": None,
            "gpu_id": None,
            "process_id": str(self.process.id),
            "extra_files": {},
            "s3_logs_put_url": None,
            "state": State.Unscheduled.value,
        })

    def test_task_definition_shm_size(self):
        self.task1.shm_size = "128g"
        self.task1.save()

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:task-definition", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.task1.id),
            "slug": "initialisation",
            "parents": [],
            "image": settings.ARKINDEX_TASKS_IMAGE,
            "command": f"python -m arkindex_tasks.init_elements {self.process.id} --chunks-number 1",
            "shm_size": "128g",
            "env": {
                "ARKINDEX_API_CSRF_COOKIE": "arkindex.csrf",
                "ARKINDEX_API_TOKEN": "deadbeefTestToken",
                "ARKINDEX_API_URL": "http://localhost:8000/api/v1/",
                "ARKINDEX_CORPUS_ID": str(self.corpus.id),
                "ARKINDEX_PROCESS_ID": str(self.process.id),
                "ARKINDEX_TASK_TOKEN": self.task1.token,
            },
            "has_docker_socket": False,
            "image_artifact_url": None,
            "agent_id": None,
            "gpu_id": None,
            "process_id": str(self.process.id),
            "extra_files": {},
            "s3_logs_put_url": None,
            "state": State.Unscheduled.value,
        })

    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.timezone")
    def test_task_final_updates_process_finished(
        self, timezone_mock, s3_mock
    ):
        """
        Updating a task to a final state should update the process's completion date
        if it causes the whole process to be finished.
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        expected_datetime = timezone.datetime(3000, 1, 1, 12).astimezone()
        timezone_mock.now.return_value = expected_datetime

        self.task1.agent = self.agent
        self.task1.state = State.Completed
        self.task1.save()
        self.task2.agent = self.agent
        self.task2.state = State.Completed
        self.task2.save()

        for state in FINAL_STATES:
            with self.subTest(state=state):
                if state == State.Stopped:
                    self.task3.state = State.Stopping
                else:
                    self.task3.state = State.Running
                self.task3.agent = self.agent
                self.task3.save()

                self.process.finished = None
                # No email is sent with mode Repository
                self.process.mode = ProcessMode.Repository
                self.process.corpus = None
                self.process.save()

                with self.assertNumQueries(11):
                    resp = self.client.patch(
                        reverse("api:task-details", args=[self.task3.id]),
                        data={"state": state.value},
                        HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                    )
                    self.assertEqual(resp.status_code, status.HTTP_200_OK)

                self.task3.refresh_from_db()
                self.assertEqual(self.task3.state, state)

                self.process.refresh_from_db()
                self.assertEqual(self.process.finished, expected_datetime)

                # No email is sent for processes of type Repository
                self.assertEqual(len(mail.outbox), 0)

    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.timezone")
    @patch("arkindex.ponos.tasks.notify_process_completion.delay")
    def test_process_failed_finished(self, async_notify_mock, timezone_mock, s3_mock):
        """
        When a task fails, its children stay unscheduled, and nothing will ever be running
        in the process anymore. It should be marked as finished.
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        expected_datetime = timezone.datetime(3000, 1, 1, 12).astimezone()
        timezone_mock.now.return_value = expected_datetime
        seal(timezone_mock)

        self.task1.state = State.Completed
        self.task3.state = State.Unscheduled
        self.task1.save()
        self.task3.save()

        for state in set(FINAL_STATES) - {State.Completed}:
            self.task2.agent = self.agent
            if state == State.Stopped:
                self.task2.state = State.Stopping
            else:
                self.task2.state = State.Running
            self.task2.save()
            with self.subTest(state=state):
                self.process.finished = None
                self.process.save()

                with self.assertNumQueries(11):
                    resp = self.client.patch(
                        reverse("api:task-details", args=[self.task2.id]),
                        data={"state": state.value},
                        HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                    )
                    self.assertEqual(resp.status_code, status.HTTP_200_OK)

                self.task2.refresh_from_db()
                self.assertEqual(self.task2.state, state)

                self.process.refresh_from_db()
                self.assertEqual(self.process.finished, expected_datetime)

    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.timezone")
    @patch("arkindex.ponos.tasks.notify_process_completion.delay")
    @override_settings(PUBLIC_HOSTNAME="http://ark.teklia/")
    def test_process_finished_email_failure(
        self, async_notify, timezone_mock, s3_mock
    ):
        async_notify.side_effect = notify_process_completion
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)
        expected_datetime = timezone.datetime(3000, 1, 1, 12).astimezone()
        timezone_mock.now.return_value = expected_datetime
        seal(timezone_mock)

        self.process.elements.set((self.page_1, self.page_2))
        self.process.activities.create(
            element=self.page_1,
            worker_version=self.dla,
            state=WorkerActivityState.Error.value,
        )
        self.process.activities.create(
            element=self.page_1,
            worker_version=self.recognizer,
            state=WorkerActivityState.Error.value,
        )

        self.process.finished = None
        self.process.name = None
        self.process.save()
        self.process.tasks.update(agent=self.agent, state=State.Completed)
        self.process.tasks.filter(id=self.task3.id).update(state=State.Running)
        with self.assertNumQueries(13):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task3.id]),
                data={"state": State.Failed.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(message.subject, f"Your process {self.process.id} finished with failures")
        self.assertListEqual(message.to, [self.process.creator.email])
        self.assertEqual(message.body, dedent(f"""
        Hello Test user,

        A process you created just finished.
        ID: {self.process.id}
        Overall state: Failed

        Tasks summary:
        * {self.task1.slug}: Completed
        * {self.task2.slug}: Completed
        * {self.task3.slug}: Failed


        Your process has 2 workers with erroneous elements:
        - Document layout analyser worker with 100% failures (1/1)

        - Recognizer worker with 100% failures (1/1)


        You can see all the details and logs here: http://ark.teklia/process/{self.process.id}/0

        --
        Arkindex

        """))

    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.timezone")
    @patch("arkindex.ponos.tasks.notify_process_completion.delay")
    @override_settings(PUBLIC_HOSTNAME="http://ark.teklia/")
    def test_process_finished_email_stopped(
        self, async_notify, timezone_mock, s3_mock
    ):
        async_notify.side_effect = notify_process_completion
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)
        expected_datetime = timezone.datetime(3000, 1, 1, 12).astimezone()
        timezone_mock.now.return_value = expected_datetime
        seal(timezone_mock)

        # Add some worker activities
        self.process.elements.set((self.page_1, self.page_2))
        self.process.activities.create(
            element=self.page_1,
            worker_version=self.dla,
            state=WorkerActivityState.Error.value,
        )
        self.process.activities.create(
            element=self.page_2,
            worker_version=self.dla,
            state=WorkerActivityState.Processed.value,
        )
        self.process.activities.create(
            element=self.page_1,
            worker_version=self.recognizer,
            state=WorkerActivityState.Queued.value,
        )

        self.process.tasks.update(agent=self.agent, state=State.Completed)
        self.process.tasks.filter(id=self.task3.id).update(state=State.Stopping)
        with self.assertNumQueries(13):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task3.id]),
                data={"state": State.Stopped.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(message.subject, f"Your process {self.process.name} finished because it was stopped")
        self.assertListEqual(message.to, [self.process.creator.email])
        self.assertEqual(message.body, dedent(f"""
        Hello Test user,

        A process you created just finished.
        ID: {self.process.id}
        Overall state: Stopped

        Tasks summary:
        * {self.task1.slug}: Completed
        * {self.task2.slug}: Completed
        * {self.task3.slug}: Stopped


        Your process has 1 worker with erroneous elements:
        - Document layout analyser worker with 50% failures (1/2)


        You can see all the details and logs here: http://ark.teklia/process/{self.process.id}/0

        --
        Arkindex

        """))

    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.timezone")
    @patch("arkindex.ponos.tasks.notify_process_completion.delay")
    @override_settings(PUBLIC_HOSTNAME="http://ark.teklia/")
    def test_process_finished_email_error(
        self, async_notify, timezone_mock, s3_mock
    ):
        async_notify.side_effect = notify_process_completion
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)
        expected_datetime = timezone.datetime(3000, 1, 1, 12).astimezone()
        timezone_mock.now.return_value = expected_datetime
        seal(timezone_mock)

        self.process.finished = None
        self.process.name = "wonderful process"
        self.process.save()
        self.process.tasks.update(agent=self.agent, state=State.Error)
        self.process.tasks.filter(id=self.task3.id).update(state=State.Running)
        with self.assertNumQueries(13):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task3.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(message.subject, "Your process wonderful process finished with errors")
        self.assertListEqual(message.to, [self.process.creator.email])
        self.assertEqual(message.body, dedent(f"""
        Hello Test user,

        A process you created just finished.
        ID: {self.process.id}
        Overall state: Error

        Tasks summary:
        * {self.task1.slug}: Error
        * {self.task2.slug}: Error
        * {self.task3.slug}: Completed

        You can see all the details and logs here: http://ark.teklia/process/{self.process.id}/0

        --
        Arkindex

        """))

    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.timezone")
    @patch("arkindex.ponos.tasks.notify_process_completion.delay")
    @override_settings(PUBLIC_HOSTNAME="http://ark.teklia/")
    def test_process_finished_email_completed_third_run(
        self, async_notify, timezone_mock, s3_mock
    ):
        async_notify.side_effect = notify_process_completion
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)
        expected_datetime = timezone.datetime(3000, 1, 1, 12).astimezone()
        timezone_mock.now.return_value = expected_datetime
        seal(timezone_mock)

        self.process.finished = None
        self.process.name = "ok"
        self.process.save()
        self.process.tasks.filter(id=self.task2.id).update(state=State.Completed, run=2)
        self.process.tasks.filter(id=self.task3.id).update(agent=self.agent, state=State.Running, run=2)
        with self.assertNumQueries(13):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task3.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(message.subject, "Your process ok finished successfully")
        self.assertListEqual(message.to, [self.process.creator.email])
        self.assertEqual(message.body, dedent(f"""
        Hello Test user,

        A process you created just finished on run 2.
        ID: {self.process.id}
        Overall state: Completed

        Tasks summary:
        * {self.task2.slug}: Completed
        * {self.task3.slug}: Completed

        You can see all the details and logs here: http://ark.teklia/process/{self.process.id}/2

        --
        Arkindex

        """))

    @patch("arkindex.ponos.signals.task_failure.send_robust")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.tasks.notify_process_completion.delay")
    def test_task_final_not_completed_triggers_signal(
        self, async_notify, s3_mock, send_mock
    ):
        """
        Updating a task to a final state that is not `State.Completed`
        should trigger the task_failure signal
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        states_expecting_signal = {State.Error, State.Failed, State.Stopped}

        for state in FINAL_STATES:
            if state is State.Completed:
                continue
            with self.subTest(state=state):
                if state is State.Stopped:
                    self.task1.state = State.Stopping
                else:
                    self.task1.state = State.Running
                self.task1.agent = self.agent
                self.task1.save()

                resp = self.client.patch(
                    reverse("api:task-details", args=[self.task1.id]),
                    data={"state": state.value},
                    HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)

                if state in states_expecting_signal:
                    self.assertEqual(send_mock.call_count, 1)
                    _, kwargs = send_mock.call_args
                    self.assertDictEqual(kwargs, {"task": self.task1})
                else:
                    self.assertFalse(send_mock.called)

                send_mock.reset_mock()

    @patch("django.contrib.auth")
    @patch("arkindex.project.aws.s3")
    def test_task_update_gpu(self, s3_mock, auth_mock):
        auth_mock.signals.user_logged_in = None
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Some nice GPU task output...")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # All GPUs are available
        self.assertFalse(GPU.objects.filter(tasks__isnull=False).exists())

        # The task is running on a GPU
        self.task1.gpu = self.agent.gpus.first()
        self.task1.agent = self.agent
        self.task1.state = State.Running
        self.task1.save()

        # One GPU is assigned
        self.assertTrue(GPU.objects.filter(tasks__isnull=False).exists())

        with self.assertNumQueries(4):
            resp = self.client.get(
                reverse("api:task-details", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        # Last ping is updated in the API call
        self.agent.refresh_from_db()

        self.assertDictEqual(
            resp.json(),
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "extra_files": {},
                "slug": "initialisation",
                "state": "running",
                "parents": [],
                "shm_size": None,
                "logs": "Some nice GPU task output...",
                "full_log": "http://somewhere",
                "agent": {
                    "cpu_cores": 2,
                    "cpu_frequency": 1000000000,
                    "cpu_load": .1,
                    "farm": {"id": str(self.agent.farm_id), "name": "Wheat farm"},
                    "gpus": [
                        {
                            "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                            "index": 0,
                            "name": "GPU1",
                            "ram_total": 2147483648,
                        },
                        {
                            "id": "f30d1407-92bb-484b-84b0-0b8bae41ca91",
                            "index": 1,
                            "name": "GPU2",
                            "ram_total": 8589934592,
                        },
                    ],
                    "hostname": "ghostname",
                    "id": str(self.agent.id),
                    "last_ping": str_date(self.agent.last_ping),
                    "ram_load": 100000000,
                    "ram_total": 2000000000,
                },
                "gpu": {
                    "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                    "index": 0,
                    "name": "GPU1",
                    "ram_total": 2147483648,
                },
            },
        )

        # Now let's complete that task, and check that the GPU is still assigned
        with self.assertNumQueries(9):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": "completed"},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            resp.json(),
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "extra_files": {},
                "slug": "initialisation",
                "state": "completed",
                "parents": [],
                "shm_size": None,
                "logs": "",
                "full_log": "http://somewhere",
                "agent": {
                    "cpu_cores": 2,
                    "cpu_frequency": 1000000000,
                    "cpu_load": .1,
                    "farm": {"id": str(self.agent.farm_id), "name": "Wheat farm"},
                    "gpus": [
                        {
                            "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                            "index": 0,
                            "name": "GPU1",
                            "ram_total": 2147483648,
                        },
                        {
                            "id": "f30d1407-92bb-484b-84b0-0b8bae41ca91",
                            "index": 1,
                            "name": "GPU2",
                            "ram_total": 8589934592,
                        },
                    ],
                    "hostname": "ghostname",
                    "id": str(self.agent.id),
                    "last_ping": str_date(self.agent.last_ping),
                    "ram_load": 100000000,
                    "ram_total": 2000000000,
                },
                "gpu": {
                    "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                    "index": 0,
                    "name": "GPU1",
                    "ram_total": 2147483648,
                },
            },
        )

        # All GPUs are available: no GPU has any task in an active state
        self.assertFalse(GPU.objects.filter(tasks__state__in=ACTIVE_STATES).exists())

    def test_update_task_from_agent_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_update_task_from_agent_forbids_users(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_task_from_agent_forbids_task(self):
        with self.assertNumQueries(1):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_task_from_agent_unassigned(self):
        self.assertNotEqual(self.task1.agent, self.agent)

        with self.assertNumQueries(2):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.ponos.models.TaskLogs.latest", new_callable=PropertyMock)
    def test_update_task_from_agent(self, short_logs_mock):
        short_logs_mock.return_value = ""
        self.process.tasks.filter(id=self.task1.id).update(state=State.Pending, agent=self.agent)
        with self.assertNumQueries(5):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Running.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        data = resp.json()
        self.assertEqual(data["id"], str(self.task1.id))
        self.assertEqual(data["state"], State.Running.value)
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Running)

    @patch("arkindex.ponos.models.TaskLogs.latest", new_callable=PropertyMock)
    def test_update_task_from_agent_completed_pends_children(self, short_logs_mock):
        """
        Child tasks whose parents are all completed get set to pending as soon as the agent updates the last parent
        """
        short_logs_mock.return_value = ""

        self.task1.agent = self.agent
        self.task2.agent = self.agent
        self.task1.state = State.Running
        self.task2.state = State.Running
        self.task1.depth = 0
        self.task2.depth = 0
        self.task3.depth = 1
        self.task1.save()
        self.task2.save()
        self.task3.save()
        self.task1.parents.clear()
        self.task2.parents.clear()
        self.task3.parents.set([self.task1, self.task2])

        with self.assertNumQueries(9):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.task2.refresh_from_db()
        self.task3.refresh_from_db()
        self.assertEqual(self.task1.state, State.Completed)
        self.assertEqual(self.task2.state, State.Running)
        # Only one of the two parents is completed: nothing happens.
        self.assertEqual(self.task3.state, State.Unscheduled)

        with self.assertNumQueries(9):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task2.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.task2.refresh_from_db()
        self.task3.refresh_from_db()
        self.assertEqual(self.task1.state, State.Completed)
        self.assertEqual(self.task2.state, State.Completed)
        # Both parents are completed, the child task is now pending
        self.assertEqual(self.task3.state, State.Pending)

    def test_update_task_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_update_task_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_task_forbids_agent(self):
        with self.assertNumQueries(0):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_update_task_forbids_task(self):
        with self.assertNumQueries(0):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_update_task_requires_process_admin_corpus(self):
        self.process.creator = self.superuser
        self.process.save()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        for role in [None, Role.Guest, Role.Contributor]:
            with self.subTest(role=role):
                self.corpus.memberships.filter(user=self.user).delete()
                if role:
                    self.corpus.memberships.create(user=self.user, level=role.value)

                with self.assertNumQueries(5):
                    resp = self.client.put(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_task_requires_process_admin_repo(self):
        self.process.mode = ProcessMode.Repository
        self.process.corpus = None
        self.process.revision = self.rev
        self.process.creator = self.superuser
        self.process.save()
        self.client.force_login(self.user)

        for role in [None, Role.Guest, Role.Contributor]:
            with self.subTest(role=role):
                self.rev.repo.memberships.filter(user=self.user).delete()
                if role:
                    self.rev.repo.memberships.create(user=self.user, level=role.value)

                with self.assertNumQueries(5):
                    resp = self.client.put(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_running_task_state_stopping(self):
        self.task1.state = State.Running
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(4):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(resp.json(), {
            "id": str(self.task1.id),
            "state": State.Stopping.value,
        })
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Stopping)

    def test_update_non_running_task_state_stopping(self):
        states = list(State)
        states.remove(State.Running)
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.put(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state {state} to state Stopping is forbidden."]},
                )

                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    def test_update_final_task_state_pending(self):
        self.task1.state = State.Completed
        self.task1.agent = self.agent
        self.task1.gpu = self.agent.gpus.first()
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Pending.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Pending)
        self.assertIsNone(self.task1.agent)
        self.assertIsNone(self.task1.gpu)

    def test_update_non_final_task_state_pending(self):
        states = set(State) - set(FINAL_STATES)
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.put(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Pending.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state {state} to state Pending is forbidden."]}
                )
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    @patch("arkindex.ponos.models.TaskLogs.latest", new_callable=PropertyMock)
    def test_partial_update_task_from_agent(self, short_logs_mock):
        short_logs_mock.return_value = ""
        self.task1.state = State.Pending
        self.task1.agent = self.agent
        self.task1.save()

        with self.assertNumQueries(5):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Running.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        data = resp.json()
        self.assertEqual(data["id"], str(self.task1.id))
        self.assertEqual(data["state"], State.Running.value)
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Running)

    def test_partial_update_task_from_agent_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_partial_update_task_from_agent_forbids_users(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_task_from_agent_forbids_task(self):
        with self.assertNumQueries(1):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_task_from_agent_unassigned(self):
        self.assertNotEqual(self.task1.agent, self.agent)

        with self.assertNumQueries(2):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.ponos.models.TaskLogs.latest", new_callable=PropertyMock)
    def test_partial_update_task_from_agent_completed_pends_children(self, short_logs_mock):
        """
        Child tasks whose parents are all completed get set to pending as soon as the agent updates the last parent
        """
        short_logs_mock.return_value = ""

        self.task1.agent = self.agent
        self.task2.agent = self.agent
        self.task1.depth = 0
        self.task2.depth = 0
        self.task3.depth = 1
        self.task1.state = State.Running
        self.task2.state = State.Running
        self.task3.state = State.Unscheduled
        self.task1.save()
        self.task2.save()
        self.task3.save()
        self.task1.parents.clear()
        self.task2.parents.clear()
        self.task3.parents.set([self.task1, self.task2])

        with self.assertNumQueries(9):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.task2.refresh_from_db()
        self.task3.refresh_from_db()
        self.assertEqual(self.task1.state, State.Completed)
        self.assertEqual(self.task2.state, State.Running)
        # Only one of the two parents is completed: nothing happens.
        self.assertEqual(self.task3.state, State.Unscheduled)
        self.assertIsNone(self.process.finished)

        with self.assertNumQueries(9):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task2.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.task2.refresh_from_db()
        self.task3.refresh_from_db()
        self.assertEqual(self.task1.state, State.Completed)
        self.assertEqual(self.task2.state, State.Completed)
        # Both parents are completed, the child task is now pending
        self.assertEqual(self.task3.state, State.Pending)
        self.assertIsNone(self.process.finished)

    def test_partial_update_task_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_partial_update_task_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_task_forbids_agent(self):
        with self.assertNumQueries(0):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_partial_update_task_forbids_task(self):
        with self.assertNumQueries(0):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_partial_update_task_requires_process_admin_corpus(self):
        self.process.creator = self.superuser
        self.process.save()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        for role in [None, Role.Guest, Role.Contributor]:
            with self.subTest(role=role):
                self.corpus.memberships.filter(user=self.user).delete()
                if role:
                    self.corpus.memberships.create(user=self.user, level=role.value)

                with self.assertNumQueries(5):
                    resp = self.client.patch(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_task_requires_process_admin_repo(self):
        self.process.mode = ProcessMode.Repository
        self.process.corpus = None
        self.process.revision = self.rev
        self.process.creator = self.superuser
        self.process.save()
        self.client.force_login(self.user)

        for role in [None, Role.Guest, Role.Contributor]:
            with self.subTest(role=role):
                self.rev.repo.memberships.filter(user=self.user).delete()
                if role:
                    self.rev.repo.memberships.create(user=self.user, level=role.value)

                with self.assertNumQueries(5):
                    resp = self.client.patch(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_running_task_state_stopping(self):
        self.task1.state = State.Running
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(4):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(resp.json(), {
            "id": str(self.task1.id),
            "state": State.Stopping.value,
        })
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Stopping)

    def test_partial_update_non_running_task_state_stopping(self):
        states = list(State)
        states.remove(State.Running)
        self.task1.save()
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.patch(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state {state} to state Stopping is forbidden."]}
                )

                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    def test_partial_update_final_task_state_pending(self):
        self.task1.state = State.Completed
        self.task1.agent = self.agent
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Pending.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Pending)
        self.assertIsNone(self.task1.agent)

    def test_partial_update_non_final_task_state_pending(self):
        states = set(State) - set(FINAL_STATES)
        self.task1.save()
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.patch(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Pending.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state {state} to state Pending is forbidden."]}
                )
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    @patch("arkindex.project.aws.s3")
    def test_task_logs_unicode_error(self, s3_mock):
        """
        Ensure the TaskLogs.latest property is able to handle sliced off Unicode characters.
        Since we fetch the latest logs from S3 using `Range: bytes=-N`, sometimes the characters
        can have a missing byte and cause a UnicodeDecodeError, such as U+00A0 (non-breaking space)
        or U+00A9 (copyright symbol).
        """
        with self.assertRaises(UnicodeDecodeError):
            b"\xa0Failed successfully".decode("utf-8")

        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"\xa0Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        self.assertEqual(data["logs"], "�Failed successfully")

    @patch("arkindex.project.aws.s3")
    def test_task_logs_errors(self, s3_mock):
        """
        Ensure the TaskLogs.latest property handles missing log files and returns an empty string,
        or handles an InvalidRange error returned only by Ceph when the log file exists but has zero bytes.
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.side_effect = ClientError({"Error": {"Code": ""}}, "get_object")
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        self.client.force_login(self.superuser)

        for code in ["NoSuchKey", "InvalidRange"]:
            with self.subTest(code=code), self.assertNumQueries(4):
                s3_mock.Object.return_value.get.side_effect.response["Error"]["Code"] = str(code)

                resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
                self.assertEqual(resp.status_code, status.HTTP_200_OK)

                self.assertEqual(resp.json()["logs"], "")

    @patch("arkindex.ponos.serializers.check_agent_key")
    @patch("arkindex.ponos.serializers.timezone")
    def test_agent_create(self, timezone_mock, check_mock):
        check_mock.return_value = True
        timezone_mock.now.return_value = timezone.datetime(2000, 1, 1, 12).astimezone()

        resp = self.client.post(
            reverse("api:agent-register"),
            {
                "hostname": "toastname",
                "cpu_cores": 42,
                "cpu_frequency": 1337e6,
                "ram_total": 16e9,
                "farm": str(self.wheat_farm.id),
                "public_key": build_public_key(),
                "derivation": "{:032x}".format(random.getrandbits(128)),
                "gpus": [
                    {
                        "id": "b6b8d7c1-c6bd-4de6-ae92-866a270be36f",
                        "name": "A",
                        "index": 0,
                        "ram_total": 512 * 1024 * 1024,
                    },
                    {
                        "id": "14fe053a-8014-46a3-ad7b-99fa4389d74c",
                        "name": "B",
                        "index": 1,
                        "ram_total": 2 * 1024 * 1024 * 1024,
                    },
                ],
            },
            format="json",
        )

        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(check_mock.call_count, 1)
        data = resp.json()
        agent = Agent.objects.get(id=data["id"])
        self.assertEqual(agent.hostname, "toastname")
        self.assertEqual(agent.cpu_cores, 42)
        self.assertEqual(agent.cpu_frequency, 1337e6)
        self.assertEqual(agent.gpus.count(), 2)
        gpu_a = agent.gpus.get(name="A")
        gpu_b = agent.gpus.get(name="B")
        self.assertEqual(gpu_a.index, 0)
        self.assertEqual(gpu_b.index, 1)
        self.assertEqual(gpu_a.ram_total, 512 * 1024 * 1024)
        self.assertEqual(gpu_b.ram_total, 2 * 1024 * 1024 * 1024)

        self.assertIn("access_token", data)
        self.assertIn("refresh_token", data)
        del data["access_token"]
        del data["refresh_token"]

        self.assertDictEqual(
            data,
            {
                "id": str(agent.id),
                "hostname": "toastname",
                "cpu_cores": 42,
                "cpu_frequency": int(1337e6),
                "farm": str(self.wheat_farm.id),
                "ram_total": 16_000_000_000,
                "cpu_load": None,
                "ram_load": None,
                "last_ping": "2000-01-01T12:00:00Z",
                "gpus": [
                    {
                        "id": "b6b8d7c1-c6bd-4de6-ae92-866a270be36f",
                        "index": 0,
                        "name": "A",
                        "ram_total": 512 * 1024 * 1024,
                    },
                    {
                        "id": "14fe053a-8014-46a3-ad7b-99fa4389d74c",
                        "index": 1,
                        "name": "B",
                        "ram_total": 2 * 1024 * 1024 * 1024,
                    },
                ],
            },
        )

    @patch("arkindex.ponos.serializers.check_agent_key")
    @patch("arkindex.ponos.serializers.timezone")
    def test_agent_update(self, timezone_mock, check_mock):
        check_mock.return_value = True
        timezone_mock.now.return_value = timezone.datetime(2000, 1, 1, 12).astimezone()

        # Still a POST, but on an existing agent
        resp = self.client.post(
            reverse("api:agent-register"),
            {
                "hostname": self.agent.hostname,
                "cpu_cores": 12,
                "cpu_frequency": 1e9,
                "ram_total": 32e9,
                "farm": str(self.wheat_farm.id),
                "public_key": self.agent.public_key,
                "derivation": "{:032x}".format(random.getrandbits(128)),
                "gpus": [
                    {
                        "id": "deadbeef-c6bd-4de6-ae92-866a270be36f",
                        "name": "new gpu",
                        "index": 0,
                        "ram_total": 32 * 1024 * 1024 * 1024,
                    },
                ],
            },
            format="json",
        )

        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(check_mock.call_count, 1)
        data = resp.json()
        self.agent.refresh_from_db()
        self.assertEqual(self.agent.hostname, "ghostname")
        self.assertEqual(self.agent.cpu_cores, 12)
        self.assertEqual(self.agent.cpu_frequency, 1e9)
        self.assertEqual(self.agent.gpus.count(), 1)
        gpu = self.agent.gpus.get(name="new gpu")
        self.assertEqual(gpu.index, 0)
        self.assertEqual(gpu.ram_total, 32 * 1024 * 1024 * 1024)
        self.assertEqual(gpu.id, uuid.UUID("deadbeef-c6bd-4de6-ae92-866a270be36f"))

        self.assertIn("access_token", data)
        self.assertIn("refresh_token", data)
        del data["access_token"]
        del data["refresh_token"]

        self.assertDictEqual(
            data,
            {
                "cpu_cores": 12,
                "cpu_frequency": 1000000000,
                "cpu_load": .1,
                "farm": str(self.wheat_farm.id),
                "gpus": [
                    {
                        "id": "deadbeef-c6bd-4de6-ae92-866a270be36f",
                        "index": 0,
                        "name": "new gpu",
                        "ram_total": 34359738368,
                    }
                ],
                "hostname": "ghostname",
                "id": str(self.agent.id),
                "last_ping": "2000-01-01T12:00:00Z",
                "ram_load": 100000000,
                "ram_total": 32000000000,
            },
        )

    @patch("arkindex.ponos.serializers.check_agent_key")
    @patch("arkindex.ponos.serializers.timezone")
    def test_agent_create_existing_gpu(self, timezone_mock, check_mock):
        check_mock.return_value = True
        timezone_mock.now.return_value = timezone.datetime(2000, 1, 1, 12).astimezone()

        resp = self.client.post(
            reverse("api:agent-register"),
            {
                "hostname": "toastname",
                "cpu_cores": 42,
                "cpu_frequency": 1337e6,
                "ram_total": 16e9,
                "farm": str(self.wheat_farm.id),
                "public_key": build_public_key(),
                "derivation": "{:032x}".format(random.getrandbits(128)),
                "gpus": [
                    {
                        "id": str(self.gpu1.id),
                        "name": "A",
                        "index": 0,
                        "ram_total": 512 * 1024 * 1024,
                    },
                    {
                        "id": str(self.gpu2.id),
                        "name": "B",
                        "index": 1,
                        "ram_total": 2 * 1024 * 1024 * 1024,
                    },
                ],
            },
            format="json",
        )
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(check_mock.call_count, 1)
        data = resp.json()

        self.gpu1.refresh_from_db()
        self.gpu2.refresh_from_db()
        self.assertEqual(self.gpu1.name, "A")
        self.assertEqual(self.gpu2.name, "B")
        self.assertEqual(self.gpu1.ram_total, 512 * 1024**2)
        self.assertEqual(self.gpu2.ram_total, 2 * 1024**3)

        new_agent = Agent.objects.get(id=data["id"])
        self.assertEqual(self.gpu1.agent, new_agent)
        self.assertEqual(self.gpu2.agent, new_agent)

        # Existing agent no longer has any assigned GPUs, since the new agent stole them
        self.assertFalse(self.agent.gpus.exists())

    @patch("arkindex.ponos.serializers.check_agent_key")
    @patch("arkindex.ponos.serializers.timezone")
    def test_agent_update_existing_gpu(self, timezone_mock, check_mock):
        check_mock.return_value = True
        timezone_mock.now.return_value = timezone.datetime(2000, 1, 1, 12).astimezone()
        self.assertTrue(self.agent.gpus.count(), 2)

        # Still a POST, but on an existing agent
        resp = self.client.post(
            reverse("api:agent-register"),
            {
                "hostname": self.agent.hostname,
                "cpu_cores": 12,
                "cpu_frequency": 1e9,
                "ram_total": 32e9,
                "farm": str(self.wheat_farm.id),
                "public_key": self.agent.public_key,
                "derivation": "{:032x}".format(random.getrandbits(128)),
                "gpus": [
                    {
                        "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                        "name": "GPU1",
                        "index": 0,
                        "ram_total": 2 * 1024 * 1024 * 1024,
                    },
                    {
                        "id": "f30d1407-92bb-484b-84b0-0b8bae41ca91",
                        "name": "GPU2",
                        "index": 1,
                        "ram_total": 8 * 1024 * 1024 * 1024,
                    },
                ],
            },
            format="json",
        )

        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        self.agent.refresh_from_db()
        self.assertTrue(self.agent.gpus.count(), 2)

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_agent_create_bad_seed(self):
        # Build keys
        server_public_key = ec.generate_private_key(
            ec.SECP384R1(),
            default_backend(),
        ).public_key()

        agent_private_key = ec.generate_private_key(
            ec.SECP384R1(),
            default_backend(),
        )
        agent_public_bytes = agent_private_key.public_key().public_bytes(
            Encoding.PEM,
            PublicFormat.SubjectPublicKeyInfo,
        )

        # Add 1 to the farm's seed
        wrong_seed = "{:064x}".format(int(self.wheat_farm.seed, 16) + 1)

        # Perform derivation with the wrong seed
        shared_key = agent_private_key.exchange(ec.ECDH(), server_public_key)
        derived_key = HKDF(
            algorithm=SHA256(),
            backend=default_backend(),
            length=32,
            salt=None,
            info=wrong_seed.encode("utf-8"),
        ).derive(shared_key)

        resp = self.client.post(
            reverse("api:agent-register"),
            {
                "hostname": "toastname",
                "cpu_cores": 42,
                "cpu_frequency": 1337e6,
                "ram_total": 16e9,
                "farm": str(self.wheat_farm.id),
                "public_key": agent_public_bytes.decode("utf-8"),
                "derivation": base64.b64encode(derived_key).decode("utf-8"),
                "gpus": [],
            },
            format="json",
        )
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            resp.json(), {"non_field_errors": ["Key verification failed"]}
        )

    def test_agent_actions_requires_token(self):
        with self.assertNumQueries(0):
            resp = self.client.get(reverse("api:agent-actions"))
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_agent_actions_no_user(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(0):
            resp = self.client.get(reverse("api:agent-actions"))
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_agent_actions_no_task(self):
        with self.assertNumQueries(0):
            resp = self.client.get(
                reverse("api:agent-actions"),
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_agent_actions_query_params_required(self):
        with self.assertNumQueries(1):
            resp = self.client.get(
                reverse("api:agent-actions"),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            resp.json(),
            {
                "cpu_load": ["This query parameter is required."],
                "ram_load": ["This query parameter is required."],
            },
        )

    def test_agent_actions_query_params_validation(self):
        with self.assertNumQueries(1):
            resp = self.client.get(
                reverse("api:agent-actions"),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                data={"cpu_load": "high", "ram_load": "low"},
            )
            self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            resp.json(),
            {
                "cpu_load": ["A valid number is required."],
                "ram_load": ["A valid number is required."],
            },
        )

    def test_agent_null_state(self):
        """
        Agents with unknown CPU or RAM load are excluded by the assignation algorithm
        """
        self.agent.cpu_load = None
        self.agent.ram_load = None
        self.agent.save()
        pubkey = build_public_key()
        second_agent = AgentUser.objects.create(
            id=hashlib.md5(pubkey.encode("utf-8")).hexdigest(),
            farm=self.wheat_farm,
            hostname="new agent",
            cpu_cores=2,
            cpu_frequency=1e9,
            public_key=pubkey,
            ram_total=1e9,
            last_ping=timezone.now(),
        )
        with self.assertNumQueries(6):
            resp = self.client.get(
                reverse("api:agent-actions"),
                HTTP_AUTHORIZATION=f"Bearer {second_agent.token.access_token}",
                data={"cpu_load": 1.9, "ram_load": 0.49},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(resp.json(), {"actions": []})

    def test_agent_non_pending_actions(self):
        """
        Only pending tasks may be retrieved as new actions
        """
        Task.objects.filter(process__farm=self.agent.farm_id).update(state=State.Error)
        with self.assertNumQueries(7):
            resp = self.client.get(
                reverse("api:agent-actions"),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                data={"cpu_load": 0.9, "ram_load": 0.49},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(resp.json(), {"actions": []})

    def test_agent_no_stealing(self):
        """
        An agent may not take another agent's tasks
        """
        Task.objects.filter(process__farm=self.agent.farm_id).update(agent=self.agent, state=State.Pending)
        pubkey = build_public_key()
        agent2 = AgentUser.objects.create(
            id=uuid.UUID(hashlib.md5(pubkey.encode("utf-8")).hexdigest()),
            farm=self.wheat_farm,
            hostname="agentorange",
            cpu_cores=2,
            cpu_frequency=1e9,
            public_key=pubkey,
            ram_total=2e9,
            last_ping=timezone.now(),
            cpu_load=.007,
        )
        with self.assertNumQueries(7):
            resp = self.client.get(
                reverse("api:agent-actions"),
                HTTP_AUTHORIZATION=f"Bearer {agent2.token.access_token}",
                data={"cpu_load": 0.9, "ram_load": 0.49},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(resp.json(), {"actions": []})

    def test_agent_actions(self):
        """
        Agent may retrieve one task using the API due to its resources limitations
        """
        self.process.farm = self.agent.farm
        self.process.save()
        self.process.tasks.update(state=State.Pending)
        now = timezone.now()

        with patch.object(api_tz, "now") as api_now_mock:
            api_now_mock.return_value = now
            with self.assertNumQueries(11):
                resp = self.client.get(
                    reverse("api:agent-actions"),
                    HTTP_AUTHORIZATION="Bearer {}".format(
                        self.agent.token.access_token
                    ),
                    data={"cpu_load": 0.9, "ram_load": 0.49},
                )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.agent.refresh_from_db()
        self.assertEqual(self.agent._estimate_new_tasks_cost(), 0.99)

        self.assertDictEqual(
            resp.json(),
            {
                "actions": [
                    {
                        "action": "start_task",
                        "task_id": str(self.task1.id),
                    }
                ]
            },
        )
        # Agent load and last ping attributes have been updated
        self.assertEqual(
            (self.agent.cpu_load, self.agent.ram_load, self.agent.last_ping),
            (0.9, 0.49, now),
        )

    def test_task_create_requires_task_auth(self):
        response = self.client.post(reverse("api:task-create"))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertDictEqual(
            response.json(),
            {"detail": "Authentication credentials were not provided."}
        )

    def test_task_create_user_forbidden(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:task-create"))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertDictEqual(
            response.json(),
            {"detail": "Authentication credentials were not provided."}
        )

    def test_task_create_agent_forbidden(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:task-create"), HTTP_AUTHORIZATION=f"Bearer {self.agent.token}")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertDictEqual(
            response.json(),
            {"detail": "Authentication credentials were not provided."}
        )

    def test_task_create_empty_body(self):
        response = self.client.post(reverse("api:task-create"), HTTP_AUTHORIZATION=f"Ponos {self.task1.token}")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "image": ["This field is required."],
                "parents": ["This field is required."],
                "slug": ["This field is required."],
            },
        )

    def test_task_create_no_parent(self):
        response = self.client.post(
            reverse("api:task-create"),
            data={
                "process_id": str(self.process.id),
                "slug": "test_task",
                "image": "registry.gitlab.com/test",
                "parents": [],
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(), {"parents": ["This list may not be empty."]}
        )

    def test_task_create_distinct_workflows_on_parents(self):
        process2 = Process.objects.create(
            farm=self.wheat_farm,
            mode=ProcessMode.Repository,
            creator=self.superuser,
        )
        task3 = process2.tasks.create(
            run=0,
            depth=1,
            slug="task_parent",
            image="registry.gitlab.com/test",
        )

        response = self.client.post(
            reverse("api:task-create"),
            data={
                "process_id": str(self.process.id),
                "slug": "test_task",
                "image": "registry.gitlab.com/test",
                "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
            },
            HTTP_AUTHORIZATION=f"Ponos {task3.token}",
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "non_field_errors": [
                    "All parents must be in the same process as the child task"
                ]
            },
        )

    def test_task_create_distinct_runs_on_parents(self):
        task3 = self.process.tasks.create(
            run=1,
            depth=1,
            slug="task_parent",
            image="registry.gitlab.com/test",
        )

        response = self.client.post(
            reverse("api:task-create"),
            data={
                "process_id": str(self.process.id),
                "slug": "test_task",
                "image": "registry.gitlab.com/test",
                "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
            },
            HTTP_AUTHORIZATION=f"Ponos {task3.token}",
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "non_field_errors": [
                    "All parents must have the same run in the given process"
                ]
            },
        )

    def test_task_create_duplicate(self):
        task = self.process.tasks.create(
            run=0,
            depth=3,
            slug="sibling",
            image="registry.gitlab.com/test",
        )

        response = self.client.post(
            reverse("api:task-create"),
            data={
                "process_id": str(self.process.id),
                "slug": "sibling",
                "image": "registry.gitlab.com/test",
                "parents": [str(self.task1.id), str(self.task2.id)],
            },
            HTTP_AUTHORIZATION=f"Ponos {task.token}",
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "non_field_errors": [
                    "A task with the `sibling` slug already exists in run 0."
                ]
            },
        )

    def test_task_create(self):
        task3 = self.process.tasks.create(
            run=0,
            depth=3,
            slug="task_parent",
            image="registry.gitlab.com/test",
        )

        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:task-create"),
                data={
                    "process_id": str(self.process.id),
                    "slug": "test_task",
                    "image": "registry.gitlab.com/test",
                    "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
                    "command": "echo Test",
                    "env": {"test": "test", "test2": "test2"},
                },
                HTTP_AUTHORIZATION=f"Ponos {task3.token}",
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_task = self.process.tasks.get(slug="test_task")

        self.assertDictEqual(
            response.json(),
            {
                "id": str(new_task.id),
                "process_id": str(self.process.id),
                "slug": "test_task",
                "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
                "image": "registry.gitlab.com/test",
                "command": "echo Test",
                "env": {
                    "ARKINDEX_API_CSRF_COOKIE": "arkindex.csrf",
                    "ARKINDEX_API_TOKEN": "deadbeefTestToken",
                    "ARKINDEX_API_URL": "http://localhost:8000/api/v1/",
                    "ARKINDEX_CORPUS_ID": str(self.process.corpus.id),
                    "ARKINDEX_PROCESS_ID": str(self.process.id),
                    "test": "test",
                    "test2": "test2",
                    "ARKINDEX_TASK_TOKEN": new_task.token,
                },
                "run": 0,
                "depth": 4,
                "has_docker_socket": False,
            },
        )

    def test_task_create_has_docker_socket_true(self):
        task3 = self.process.tasks.create(
            run=0,
            depth=3,
            slug="task_parent",
            image="registry.gitlab.com/test",
        )

        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:task-create"),
                data={
                    "process_id": str(self.process.id),
                    "slug": "test_task",
                    "image": "registry.gitlab.com/test",
                    "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
                    "command": "echo Test",
                    "env": {"test": "test", "test2": "test2"},
                    "has_docker_socket": True,
                },
                HTTP_AUTHORIZATION=f"Ponos {task3.token}",
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_task = self.process.tasks.get(slug="test_task")

        self.assertDictEqual(
            response.json(),
            {
                "id": str(new_task.id),
                "process_id": str(self.process.id),
                "slug": "test_task",
                "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
                "image": "registry.gitlab.com/test",
                "command": "echo Test",
                "env": {
                    "ARKINDEX_API_CSRF_COOKIE": "arkindex.csrf",
                    "ARKINDEX_API_TOKEN": "deadbeefTestToken",
                    "ARKINDEX_API_URL": "http://localhost:8000/api/v1/",
                    "ARKINDEX_CORPUS_ID": str(self.process.corpus.id),
                    "ARKINDEX_PROCESS_ID": str(self.process.id),
                    "test": "test",
                    "test2": "test2",
                    "ARKINDEX_TASK_TOKEN": new_task.token,
                },
                "run": 0,
                "depth": 4,
                "has_docker_socket": True,
            },
        )

    def test_retrieve_secret_requires_auth(self):
        """
        Only agents or tasks may access a secret details
        """
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:secret-details", kwargs={"name": "abc"})
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertDictEqual(
            response.json(), {"detail": "Authentication credentials were not provided."}
        )

    def test_retrieve_secret_not_found(self):
        """
        A 404 should be raised when no secret match query name
        """
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse(
                    "api:secret-details", kwargs={"name": "the_most_important_secret"}
                ),
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_retrieve_secret_agent_forbidden(self):
        """
        Agent should be able to retrieve a secret content as cleartext
        """
        account_name = "bank_account/0001/private"
        secret = Secret.objects.create(
            name=account_name,
            nonce=b"1337" * 4,
            content=encrypt(b"1337" * 4, "1337$"),
        )
        self.assertEqual(secret.content, b"\xc1\x81\xc0\xceo")

        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:secret-details", kwargs={"name": account_name}),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_retrieve_secret_forbid_users(self):
        account_name = "bank_account/0001/private"
        secret = Secret.objects.create(
            name=account_name,
            nonce=b"1337" * 4,
            content=encrypt(b"1337" * 4, "1337$"),
        )
        self.assertEqual(secret.content, b"\xc1\x81\xc0\xceo")

        for user in [self.user, self.superuser]:
            with self.subTest(user=user):
                self.client.force_login(user)
                with self.assertNumQueries(0):
                    response = self.client.get(reverse("api:secret-details", kwargs={"name": account_name}))
                    self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_task_retrieve_secret_requires_active_user(self):
        account_name = "bank_account/0001/private"
        secret = Secret.objects.create(
            name=account_name,
            nonce=b"1337" * 4,
            content=encrypt(b"1337" * 4, "1337$"),
        )
        self.assertEqual(secret.content, b"\xc1\x81\xc0\xceo")

        Process.objects.create(
            creator=self.user,
            mode=ProcessMode.Repository,
        )
        self.user.is_active = False
        self.user.save()

        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:secret-details", kwargs={"name": account_name}),
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

        self.assertDictEqual(
            response.json(),
            {"detail": "User inactive or deleted."},
        )

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_task_retrieve_secret(self):
        account_name = "bank_account/0001/private"
        secret = Secret.objects.create(
            name=account_name,
            nonce=b"1337" * 4,
            content=encrypt(b"1337" * 4, "1337$"),
        )
        self.assertEqual(secret.content, b"\xc1\x81\xc0\xceo")

        corpus = Corpus.objects.create(name="Test corpus")
        corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=self.user,
        )

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:secret-details", kwargs={"name": account_name}),
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            response.json(),
            {
                "id": str(secret.id),
                "name": account_name,
                "content": "1337$",
            },
        )

    def test_list_agents_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:agents-state"))
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_agents_agent_forbidden(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:agents-state"),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_agents_task_forbidden(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:agents-state"),
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_agents_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:agents-state"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_agents(self):
        """
        Lists agents from all farms with their status
        """
        # Add tasks with different states on the agent
        self.agent.tasks.bulk_create(
            [
                Task(
                    run=0,
                    depth=0,
                    process=self.process,
                    slug=state.value,
                    state=state,
                    agent=self.agent,
                )
                for state in State
            ]
        )
        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:agents-state"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 1)
        agent_state = data["results"][0]
        del agent_state["last_ping"]
        self.assertDictEqual(
            agent_state,
            {
                "id": str(self.agent.id),
                "active": True,
                "cpu_cores": 2,
                "cpu_frequency": 1000000000,
                "cpu_load": .1,
                "farm": {
                    "id": str(self.wheat_farm.id),
                    "name": "Wheat farm",
                },
                "gpus": [
                    {
                        "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                        "index": 0,
                        "name": "GPU1",
                        "ram_total": 2147483648,
                    },
                    {
                        "id": "f30d1407-92bb-484b-84b0-0b8bae41ca91",
                        "index": 1,
                        "name": "GPU2",
                        "ram_total": 8589934592,
                    },
                ],
                "hostname": "ghostname",
                "ram_load": 100000000,
                "ram_total": 2000000000,
                "running_tasks_count": 1,
            },
        )

    def test_retrieve_agent_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:agent-details", kwargs={"pk": str(self.agent.id)}),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_retrieve_agent_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:agent-details", kwargs={"pk": str(self.agent.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_agent_details(self):
        """
        The view returns an agents with its details and associated running tasks
        """
        # Add tasks with different states on the agent
        Task.objects.bulk_create(
            [
                Task(
                    run=0,
                    depth=0,
                    process=self.process,
                    slug=state.value,
                    state=state,
                    agent=self.agent,
                )
                for state in State
            ]
        )
        running_task = self.agent.tasks.get(state=State.Running)
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:agent-details", kwargs={"pk": str(self.agent.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        del data["last_ping"]
        self.assertDictEqual(
            data,
            {
                "id": str(self.agent.id),
                "active": True,
                "cpu_cores": 2,
                "cpu_frequency": 1000000000,
                "cpu_load": .1,
                "farm": {
                    "id": str(self.wheat_farm.id),
                    "name": "Wheat farm",
                },
                "gpus": [
                    {
                        "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                        "index": 0,
                        "name": "GPU1",
                        "ram_total": 2147483648,
                    },
                    {
                        "id": "f30d1407-92bb-484b-84b0-0b8bae41ca91",
                        "index": 1,
                        "name": "GPU2",
                        "ram_total": 8589934592,
                    },
                ],
                "hostname": "ghostname",
                "ram_load": 100000000,
                "ram_total": 2000000000,
                "running_tasks": [
                    {
                        "id": str(running_task.id),
                        "run": 0,
                        "depth": 0,
                        "parents": [],
                        "slug": "running",
                        "state": "running",
                        "shm_size": None,
                    }
                ],
            },
        )

    def test_retrieve_agent_agent_forbidden(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:agent-details", kwargs={"pk": str(self.agent.id)}),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_retrieve_agent_task_forbidden(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:agent-details", kwargs={"pk": str(self.agent.id)}),
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_farms_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:farm-list"))
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_farms_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:farm-list"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_farms_agent_forbidden(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:farm-list"), HTTP_AUTHORIZATION=f"Bearer {self.agent.token}")
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_farms_task_forbidden(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:farm-list"), HTTP_AUTHORIZATION=f"Ponos {self.task1.token}")
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_farms(self):
        self.client.force_login(self.superuser)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:farm-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            response.json(),
            {
                "count": 2,
                "number": 1,
                "previous": None,
                "next": None,
                "results": [
                    {"id": str(self.default_farm.id), "name": "Default farm"},
                    {"id": str(self.wheat_farm.id), "name": "Wheat farm"},
                ],
            },
        )

    def test_list_farms_guest(self):
        self.client.force_login(self.user)
        self.assertFalse(self.default_farm.is_available(self.user))
        self.assertTrue(self.wheat_farm.is_available(self.user))
        # Accessing memberships causes content types to be cached,
        # so clear again to get the amount of queries for the worst case
        self.clear_caches()

        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:farm-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            response.json(),
            {
                "count": 1,
                "number": 1,
                "previous": None,
                "next": None,
                "results": [
                    {"id": str(self.wheat_farm.id), "name": "Wheat farm"},
                ],
            },
        )

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_public_key(self):
        expected = load_pem_private_key(
            PONOS_PRIVATE_KEY.read_bytes(),
            password=None,
            backend=default_backend(),
        ).public_key().public_bytes(
            Encoding.PEM,
            PublicFormat.SubjectPublicKeyInfo,
        )

        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:public-key"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.content, expected)

    def test_agent_refresh(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:agent-token-refresh"),
                {"refresh": str(self.agent.token)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertCountEqual(list(data.keys()), ["access", "refresh"])
        access_token = AccessToken(data["access"])
        refresh_token = RefreshToken(data["refresh"])
        access_token.verify()
        refresh_token.verify()
        self.assertEqual(access_token.payload["agent_id"], str(self.agent.id))
        self.assertEqual(refresh_token.payload["agent_id"], str(self.agent.id))
